-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================


library ieee;
use ieee.std_logic_1164.all;	   
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use work.sha3_blake_package.all;
            
entity permute4xor_ppl is
	generic ( 		 
	ppl : integer := 4;
	h : integer := 256;
	b : integer := 512;	
	iw : integer := 32 );
	port(		
		clk 		: in std_logic;		
		m			:	in std_logic_vector(b-1 downto 0);
		round		:	in std_logic_vector(4 downto 0);
		consout		:	out std_logic_vector(b/2-1 downto 0)
	);
end permute4xor_ppl;	

architecture muxbased of permute4xor_ppl is	 
	type std_logic_matrix is array (15 downto 0) of std_logic_vector(iw - 1 downto 0) ;
	type std_logic_half_matrix is array (7 downto 0) of std_logic_vector(iw - 1 downto 0) ;
	type ctmp_type is array (3 downto 0) of std_logic_vector(iw - 1 downto 0) ;
	signal ctmp1, ctmp2 :  ctmp_type;
	--------------------------
	function wordmatrix2halfblk  	(x : std_logic_half_matrix) return std_logic_vector is
		variable retval : std_logic_vector(b/2-1 downto 0);
	begin
		for i in 0 to 7 loop
			retval(iw*(i+1) - 1 downto iw*i) := x(7-i);
		end loop;
		return retval;
	end wordmatrix2halfblk;
	--------------------------
	function blk2wordmatrix  	(x : std_logic_vector ) return std_logic_matrix is
		variable retval : std_logic_matrix;
	begin
		for i in 0 to 15 loop
			retval(15-i) := x(iw*(i+1) - 1 downto iw*i);
		end loop;
		return retval;
	end blk2wordmatrix;
	-------------------------- 
	
	signal mblk : std_logic_matrix;	
	signal minblkprime : std_logic_half_matrix;

	type block_array is array(0 to 19) of std_logic_half_matrix;
	signal mblkprime : block_array; 	
	signal mprime_tmp, consprime_tmp, consout_tmp : std_logic_half_matrix;	
	
	function get_halfmatrixzero return std_logic_half_matrix is
		variable ret : std_logic_half_matrix;
	begin
		for i in 0 to 7 loop
			ret(0) := (others => '0');
		end loop;
		return ret;
	end function get_halfmatrixzero;  
	constant zero : std_logic_half_matrix := get_halfmatrixzero;
	--------------------------								   
	constant consin : std_logic_vector(b-1 downto 0) := get_cons( h, b, iw );
	function get_cp ( gsize : integer; iw : integer ) return block_array is
		variable cblk : std_logic_matrix;   									   		
		variable cpblk : block_array;	   
								  
	begin							   
		for i in 0 to 15 loop
			cblk(15-i) := consin(iw*(i+1) - 1 downto iw*i);
		end loop;
		for i in 0 to 9 loop
			for j in 0 to 16/(8/gsize)-1 loop
				cpblk(2*i)(j) 	:= cblk( permute_array( i, j ) );
				cpblk(2*i+1)(j) := cblk( permute_array( i, j+8 ) );
			end loop;
		end loop;		
		return cpblk;
	end function get_cp;		
	------------------------------
	
	constant consblkprime : block_array := get_cp( 4, iw );	  
begin		
	mblk <= blk2wordmatrix( m );
	
	ret1_gen : for i in 0 to 9 generate
		ret2_gen : for j in 0 to 7 generate	 	
			mblkprime(2*i)(j) 		<= mblk( permute_array( i, j ) );
			mblkprime(2*i+1)(j) 	<= mblk( permute_array( i, j+8 ) );
		end generate;
	end generate; 

	with round select 
	mprime_tmp	 <= 	mblkprime(0) 	when "00000",
						mblkprime(1)  	when "00001",
						mblkprime(2)  	when "00010",
						mblkprime(3)  	when "00011",
						mblkprime(4)  	when "00100",
						mblkprime(5)  	when "00101",
						mblkprime(6)  	when "00110",
						mblkprime(7)  	when "00111",
						mblkprime(8)  	when "01000",
						mblkprime(9)   when "01001",
						mblkprime(10)  when "01010",
						mblkprime(11)  when "01011",
						mblkprime(12)  when "01100",	
						mblkprime(13)  when "01101",	
						mblkprime(14)  when "01110",	
						mblkprime(15)  when "01111",	
						mblkprime(16)  when "10000",	
						mblkprime(17)  when "10001",	
						mblkprime(18)  when "10010",	
						mblkprime(19)   when "10011",	
						mblkprime(0)   when "10100",	
						mblkprime(1)   when "10101",	
						mblkprime(2)   when "10110",	
						mblkprime(3)   when "10111",	
						mblkprime(4)   when "11000",	
						mblkprime(5)   when "11001",	
						mblkprime(6)   when "11010",
						mblkprime(7)  	when "11011",
						mblkprime(8)  	when "11100",
						mblkprime(9)  	when "11101",
						mblkprime(10) 	when "11110",
						mblkprime(11) 	when "11111",						
						zero when  others ;
	
	with round select 
	consprime_tmp	 <= 	consblkprime(0) 	when "00000",
					consblkprime(1)  	when "00001",
					consblkprime(2)  	when "00010",
					consblkprime(3)  	when "00011",
					consblkprime(4)  	when "00100",
					consblkprime(5)  	when "00101",
					consblkprime(6)  	when "00110",
					consblkprime(7)  	when "00111",
					consblkprime(8)  	when "01000",
					consblkprime(9)   when "01001",
					consblkprime(10)  when "01010",
					consblkprime(11)  when "01011",
					consblkprime(12)  when "01100",	
					consblkprime(13)  when "01101",	
					consblkprime(14)  when "01110",	
					consblkprime(15)  when "01111",	
					consblkprime(16)  when "10000",	
					consblkprime(17)  when "10001",	
					consblkprime(18)  when "10010",	
					consblkprime(19)   when "10011",	
					consblkprime(0)   when "10100",	
					consblkprime(1)   when "10101",	
					consblkprime(2)   when "10110",	
					consblkprime(3)   when "10111",	
					consblkprime(4)   when "11000",	
					consblkprime(5)   when "11001",	
					consblkprime(6)   when "11010",
					consblkprime(7)  	when "11011",
					consblkprime(8)  	when "11100",
					consblkprime(9)  	when "11101",
					consblkprime(10) 	when "11110",
					consblkprime(11) 	when "11111",
					zero	when  others ;
				
	output_gen : for i in 0 to 3 generate
		consout_tmp(i*2) 	<= mprime_tmp(i*2) xor consprime_tmp(i*2+1);
		consout_tmp(i*2+1) 	<= mprime_tmp(i*2+1) xor consprime_tmp(i*2);
	end generate;
	
	outputreg : process ( clk )
	begin
		if rising_edge( clk ) then	 	
			-- Shift register to delay the permute output for 2 additional clock cycles
			if ppl = 4 then
				for i in 0 to 3 loop
					ctmp1(i) <= consout_tmp(2*i+1);
					ctmp2(i) <= ctmp1(i);
				end loop;
				consout <= 	consout_tmp(0) & ctmp2(0) & consout_tmp(2) & ctmp2(1) & 
							consout_tmp(4) & ctmp2(2) & consout_tmp(6) & ctmp2(3);
			end if;			
			if ppl = 2 then
				consout <= 	consout_tmp(0) & consout_tmp(1) & consout_tmp(2) & consout_tmp(3) & 
							consout_tmp(4) & consout_tmp(5) & consout_tmp(6) & consout_tmp(7);
			end if;
		end if;
	end process;  
end muxbased;	

-------------------------------------------
-------------------------------------------
-------------------------------------------
-------------------------------------------	 


library ieee;
use ieee.std_logic_1164.all;	   
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use work.sha3_blake_package.all;
            
entity permute8xor_ppl is
	generic (		
	ppl : integer := 2 ;
	h : integer := 256;
	b : integer := 512;
	iw : integer := 32 );
	port(			  	
		clk : in std_logic;		
		m		:	in std_logic_vector(b-1 downto 0);
		round		:	in std_logic_vector(3 downto 0);
		consout		:	out std_logic_vector(b-1 downto 0)
	);
end permute8xor_ppl;		  

architecture muxbased of permute8xor_ppl is	 	 
	type ctmp_type is array (16/ppl-1 downto 0) of std_logic_vector(iw - 1 downto 0) ;
	signal cmtmp1, cmtmp2_1, cmtmp2_2, cmtmp3_1, cmtmp3_2, cmtmp3_3 :  ctmp_type;

	type std_logic_matrix is array (15 downto 0) of std_logic_vector(iw - 1 downto 0) ;
	function wordmatrix2blk  	(x : std_logic_matrix) return std_logic_vector is
		variable retval : std_logic_vector(b-1 downto 0);
	begin
		for i in 0 to 15 loop
			retval(iw*(i+1) - 1 downto iw*i) := x(15-i);
		end loop;
		return retval;
	end wordmatrix2blk;
	function blk2wordmatrix  	(x : std_logic_vector ) return std_logic_matrix is
		variable retval : std_logic_matrix;
	begin
		for i in 0 to 15 loop
		retval(15-i) := x(iw*(i+1) - 1 downto iw*i);
		end loop;
		return retval;
	end blk2wordmatrix;		  
	
	constant consin : std_logic_vector(b-1 downto 0) := get_cons( h, b, iw );
	
	signal mblk, consblk  : std_logic_matrix;
	type block_array is array(0 to 9) of std_logic_matrix;
	signal mblkprime, consblkprime : block_array; 
	signal mprime_tmp, consprime_tmp, cmtmp : std_logic_matrix;	
	
begin		
	mblk <= blk2wordmatrix( m );
	consblk <= blk2wordmatrix( consin );
	
	ret1_gen : for i in 0 to 9 generate
		ret2_gen : for j in 0 to 15 generate	 	
			mblkprime(i)(j) 	<= mblk( permute_array( i, j ) );
			consblkprime(i)(j) 	<= consblk( permute_array( i, j ) );
		end generate;
	end generate; 

	with round select 
	mprime_tmp	 <= mblkprime(0) when "0000",
					mblkprime(1) when "0001",
					mblkprime(2) when "0010",
					mblkprime(3) when "0011",
					mblkprime(4) when "0100",
					mblkprime(5) when "0101",
					mblkprime(6) when "0110",
					mblkprime(7) when "0111",
					mblkprime(8) when "1000",
					mblkprime(9) when "1001",
					mblkprime(0) when "1010",
					mblkprime(1) when "1011",
					mblkprime(2) when "1100",
					mblkprime(3) when "1101",	
					mblkprime(4) when "1110",	
					mblkprime(5) when others;

	with round select 
	consprime_tmp	 <= consblkprime(0) when "0000",
					consblkprime(1) when "0001",
					consblkprime(2) when "0010",
					consblkprime(3) when "0011",
					consblkprime(4) when "0100",
					consblkprime(5) when "0101",
					consblkprime(6) when "0110",
					consblkprime(7) when "0111",
					consblkprime(8) when "1000",
					consblkprime(9) when "1001",
					consblkprime(0) when "1010",
					consblkprime(1) when "1011",
					consblkprime(2) when "1100",
					consblkprime(3) when "1101",	
					consblkprime(4) when "1110",	
					consblkprime(5) when others;
	
	
	output_gen : for i in 0 to 7 generate
		cmtmp(i*2) 		<= mprime_tmp(i*2) xor consprime_tmp(i*2+1);
		cmtmp(i*2+1) 	<= mprime_tmp(i*2+1) xor consprime_tmp(i*2);
	end generate;
	
	
	anotherinreg : process ( clk )
	begin
		if rising_edge( clk ) then	 												
			if ppl = 4 then		 
				for i in 0 to 3 loop
					cmtmp1(i) <= cmtmp(2*i+1);
					cmtmp2_1(i) <= cmtmp(8+2*i);
					cmtmp3_1(i) <= cmtmp(8+2*i+1);
				end loop;						  
				cmtmp2_2 <= cmtmp2_1;
				cmtmp3_2 <= cmtmp3_1;
				cmtmp3_3 <= cmtmp3_2;
				consout <=	cmtmp(0) & cmtmp1(0) & cmtmp(2) & cmtmp1(1) & 
							cmtmp(4) & cmtmp1(2) & cmtmp(6) & cmtmp1(3) & 
							cmtmp2_2(0) & cmtmp3_3(0) & cmtmp2_2(1) & cmtmp3_3(1) & 
							cmtmp2_2(2) & cmtmp3_3(2) & cmtmp2_2(3) & cmtmp3_3(3);
			end if;
			
			if ppl = 2 then		 
				for i in 0 to 7 loop
					cmtmp1(i) <= cmtmp(i+8);
				end loop;
				consout <=	cmtmp(0) & cmtmp(1) & cmtmp(2) & cmtmp(3) & 
							cmtmp(4) & cmtmp(5) & cmtmp(6) & cmtmp(7) & 
							cmtmp1(0) & cmtmp1(1) & cmtmp1(2) & cmtmp1(3) & 
							cmtmp1(4) & cmtmp1(5) & cmtmp1(6) & cmtmp1(7);
			end if;
			
		end if;
	end process;  
	
	
end muxbased;	